#!/bin/env -S deno run --allow-read

async function main() {
	const { path } = parse_args(Deno.args);
	const content = await Deno.readTextFile(path);
	const { targets, definitions } = parse_file(content);
	const ingredients = calculate_ingredients(targets, definitions);
	const merged = merge_ingredients(ingredients);
	console.log("To produce :\n", ...format_ingredients(targets), "\nGet :\n", ...format_ingredients(merged));
}

function parse_args(args: string[]) {
	const invalid_usage = () => Deno.exit(console.log("Usage:\nreciper <recipe_file>") as undefined);
	if (args.length !== 1) invalid_usage();
	const [path] = args;
	if (["-h", "--help"].includes(path)) invalid_usage();
	return { path };
}

function parse_file(content: string) {
	const lines = content.trim().split("\n");
	const definitions = [] as Definition[];
	const main = { target: "main", ingredients: [] } as Definition;
	let current_block = null as Definition | null;
	for (const line of lines) {
		if (line.trim().startsWith("#")) continue;
		if (line.trim() === "") {
			if (current_block !== null) definitions.push(current_block);
			current_block = null;
			continue;
		}
		const [count_str, ...words] = line.split(" ");
		const count = parseFloat(count_str);
		if (isNaN(count)) {
			if (current_block !== null) definitions.push(current_block);
			current_block = null;
			current_block = { target: line, ingredients: [] };
		} else {
			const item = words.join(" ");
			if (current_block === null) main.ingredients.push({ item, count });
			else current_block.ingredients.push({ item, count });
		}
	}
	if (current_block !== null) definitions.push(current_block);
	const targets = main.ingredients;
	return { targets, definitions };
}

type Ingredient = { item: string; count: number };
type Definition = { target: string; ingredients: Ingredient[] };

function calculate_ingredients(targets: Ingredient[], definitions: Definition[]) {
	const result = [] as Ingredient[];
	for (const target of targets) {
		const recipe = definitions.find((recipe) => recipe.target === target.item);
		if (recipe === undefined) {
			result.push(target);
			continue;
		}
		const defs = definitions.filter((def) => def.target !== target.item);
		for (const ingredient of calculate_ingredients(recipe.ingredients, defs)) {
			const count = ingredient.count * target.count;
			result.push({ item: ingredient.item, count });
		}
	}
	return result;
}

function merge_ingredients(ingredients: Ingredient[]): Ingredient[] {
	const set = new Map<string, number>();
	for (const { count, item } of ingredients) set.set(item, count + (set.get(item) ?? 0));
	return Array.from(set.entries()).map(([item, count]) => ({ item, count }));
}

function* format_ingredients(ingredients: Ingredient[]) {
	for (const { count, item } of ingredients) yield* ["-", count, item, "\n"];
}

if (import.meta.main) await main();
